/**
% The GraphicalSimulationEngine is a tool that provides a common interface
% to a simulation engine, simulating message passing between nodes over links,
% and two graphical representation of the events in that simulation engine.
%
% A precise definition of the behaviour of the simulation engine can be found
% in the documentation of the SimulationEngine functor and class.
% The two graphical representations are NodeLink and MessageLog. NodeLink
% shows a graphical snapshot of nodes, links and messages present in the
% Simulation at any instant. MessageLog shows a graphical log of all messages
% sent and received by each node. Both are available as frames.
%
% For the NodeLink and MessageLog tools to work messages must fit the
% definition of class Message.
*/
functor
export
   GraphicalSimulationEngine
   Message
import
   SimulationEngine at '../SimulationEngine/SimulationEngine.ozf'
   NodeLinkGraphics at '../NodeLinkGraphics/NodeLinkGraphics.ozf'
   MessageLog at '../MessageLog/MessageLog.ozf'
define
   class GraphicalSimulationEngine
      % If step happens while e.g. the SimulationEngine has a link
      % but NodeLinkGraphics does not yet have it, information is inconsistent
      % therefore use locks.
      prop
	 locking
      feat
	 simulationEngine
	 nodeLinkGraphics
	 messageLog
	 Links
      attr
	 nodeLinkActive
	 messageLogActive
	 delays:nil
	 

      /**
      % @arg 1 NodeLinkTopLevel parent for the Node Layout-frame
      % @arg 2 MessageLogTopLevel parent for the Message Log-frame
      % @arg 3 NodeLinkFrame returned frame with Node Layout
      % @arg 4 MessageLogFrame returned frame with Message Log
      */
      meth init(NodeLinkTopLevel MessageLogTopLevel
		?NodeLinkFrame ?MessageLogFrame)
	 self.simulationEngine={New SimulationEngine.simulationEngine init}
	 self.nodeLinkGraphics={New NodeLinkGraphics.nodeLinkGraphics
				init(NodeLinkTopLevel NodeLinkFrame)}
	 self.messageLog={New MessageLog.messageLog
			  init(MessageLogTopLevel MessageLogFrame)}
	 self.Links={NewDictionary}
	 nodeLinkActive<-true
	 messageLogActive<-true
      end

      /**
      % Advances the simulation one step. This means:
      % 1. do step on the SimulationEngine
      % 2. if active tell NodeLinkGraphics the current state with step(...)
      % 3. if active tell MessageLog to calculate and display its graph
      */
      meth step
	 lock
	    {self.simulationEngine step}
	    if @nodeLinkActive then
	       {self.nodeLinkGraphics step({self.simulationEngine getTime($)}
					   {self.simulationEngine
					    getLinkStates($)})}
	    end
	    if @messageLogActive then
	       {self.messageLog solve}
	    end
	    delays <- {Unwind {self getTime($)} @delays}
	 end
      end

      /**
      % Set the NodeLinkGraphics be active or inactive with true or false.
      % When inactive, it will receive information on created/removed
      % nodes/links, but won't update its state on the links.
      */
      meth nodeLinkActive(Bool)
	 nodeLinkActive<-Bool
	 if Bool then % Update now!
	    lock
	       {self.nodeLinkGraphics step({self.simulationEngine getTime($)}
					   {self.simulationEngine
					    getLinkStates($)})}
	    end
	 end
      end

      /**
      % Set the MessageLog be active or inactive with true or false.
      % When inactive, it will receive send and receive events but won't
      % calculate any new state graphs.
      */
      meth messageLogActive(Bool)
	 messageLogActive<-Bool
	 if Bool then % Update now!
	    lock
	       {self.messageLog solve}
	    end
	 end
      end

      /**
      % Create a new node with the specified id and listener. Further
      % description of this part is found under SimulationEngine.
      % A specification of coordinates or color may be given in GraphSpec
      % as documented in NodeLinkGraphics.
      */
      meth createNode(NodeId Listener GraphSpec<=unit)
	 % Wrap Listener to be able to notify message log of received messages
	 proc{WrapListener M}
	    case M of step(_ Msgs) then
	       {ForAll Msgs proc{$ Msg#From}
			       if Msg\=connectionLost then
				  {self.messageLog receive('from':From
							   to:NodeId
							   msg:Msg)}
			       end
			    end}
	    end
	    {Listener M}
	 end
      in
	 lock
	    {self.simulationEngine createNode(NodeId WrapListener)}
	    {self.nodeLinkGraphics createNode(NodeId GraphSpec)}
	    {self.messageLog createNode(NodeId
					{self.nodeLinkGraphics
					 getColorOfNode(NodeId $)})}
	 end
      end

      /**
      % Remove node nodeid. Behaviour specified under SimulationEngine.
      */
      meth removeNode(NodeId)
	 lock
	    {self.simulationEngine removeNode(NodeId)}
	    {self.nodeLinkGraphics removeNode(NodeId)}
	    % messageLog remove?
	 end
      end

      /**
      % Create a new link between the two specified nodes. A link identifier
      % to be used at sends and removals is returned in L.
      % Optionally the number of slots in the link and the shifter can be
      % specified. Further see SimulationEngine.
      */
      meth createLink(FromNodeId ToNodeId ?L
		      nofslots:NofSlots<=auto shifter:Shifter<=default) N in
	 lock
	    L = {NewName}
	    self.Links.L:=FromNodeId#ToNodeId % save for use with message log
	    if NofSlots==auto then
	       N={self.nodeLinkGraphics calcNofSlots(FromNodeId ToNodeId $)}
	    else
	       N=NofSlots
	    end
	    
	    {self.simulationEngine createLink(L FromNodeId ToNodeId N Shifter)}
	    {self.nodeLinkGraphics createLink(L FromNodeId ToNodeId N)}
	 end
      end

      /**
      % Remove the link in one of two ways. Default is crash wich will
      % remove the link immediately dropping any messages in the link.
      % Optionally flush can be specified instead. This will remove the
      % link after all messages have traveled through the link. The link
      % is immediately closed for incomming messages. A flushing link state
      % is known to the NodeLinkGraphics enabling NodeLinkGraphics to display
      % a flushing link in e.g. a different color.
      */
      meth removeLink(L How<=crash)
	 proc{DoneAlert}
	    {self.nodeLinkGraphics removeLink(L)}
	    % messageLog remove?
	    {Dictionary.remove self.Links L}
	 end
      in
	 lock
	    {self.simulationEngine removeLink(L How DoneAlert)}
	 end
      end

      /**
      % send a message Msg on link L. Message must look like specified by
      % class Message.
      % If several threads to simultaneous sends ont the same link, it is
      % not sure that they will appear in the same order in the
      % SimulationEngine as in the MessageLog.
      */
      meth send(L Msg)
	 {self.simulationEngine send(L Msg)}
	 local From#To=self.Links.L in
	    {self.messageLog send(to:To 'from':From msg:Msg)}
	 end
      end

      /**
      % returns the current time of the engine.
      */
      meth getTime($)
	 {self.simulationEngine getTime($)}
      end

      /**
      % Update the text shown in a node. Every entry in TextList will
      % be shown on a seperate line.
      */
      meth updateNodeText(NodeId TextList)
	 {self.nodeLinkGraphics updateNodeText(NodeId TextList)}
      end
	 
      /**
      % insert a string in the message log at node Id
      */
      meth insertLogString(Id S)
	 {self.messageLog string('from':Id msg:S)}
      end

      /**
      % level the message log
      */
      meth levelLog
	 {self.messageLog level}
      end

      /**
      % Perform a delay based on simulation time
      */
      meth delay(Time) A O N in
	 O=delays<-N
	 N={InsertDelay {self getTime($)} + Time A O}
	 {Wait A}
      end
   end

   /**
   % This class may be viewed as a template. All messages that are sent
   % on the graphical simulation engine must have at least the following
   % interface.
   % The graphical simulation engine will never alter messages. It does not
   % care what contents contains. id must be a unique identifier.
   */
   class Message
      feat
	 id
	 content

      meth init(Content)
	 self.id={NewName}
	 self.content=Content
      end

      meth toString($)
	 {VirtualString.toAtom {Value.toVirtualString self.content 1000 1000}}
      end

      /**
      % Implement this method to return e.g. the message type as a string or
      % atom.
      */
      meth toShortString($)
	 {self toString($)}
      end

      /**
      % Override this method to give some or all messages a different color
      % (Should return a Tk valid color)
      */
      meth getMessageColor($)
	 black
      end

      /**
      % Override this method to give some or all messages a different shape
      % (Should return one of the following shapes: square, circle, triangle or
      % hexagon)
      */
      meth getMessageShape($)
	 circle
      end

   end

   fun{Unwind Time L}
      case L of nil then nil
      elseof !Time#V|R then
	 V = unit
	 {Unwind Time R}
      else L end
   end
   
   fun{InsertDelay Time Sync L}
      case L of nil then [Time#Sync]
      elseof T#S|R then
	 if T>Time then Time#Sync|L
	 else T#S|{InsertDelay Time Sync R} end
      end
   end

   % It is convenient with a human readable name for debugging
%\define USE_ALT_NEWNAME
\ifdef USE_ALT_NEWNAME
   Next={NewCell 0}
   fun{NewName} O N in
      {Exchange Next O N}
      N=O+1
      O
   end
\endif
end
