/**
% The exported class implements a remote control that optionally can be used
% with the GraphicalSimulationEngine. The remote control has one graphical
% interface represented by a frame to be shown as desired, and one method
% interface that can be invoked from the OPI or other compiled code.
% Two toplevels are expected, one containing the NodeLinkFrame and one
% containing the MessageLog. These can be opened and closed with use of the
% remote control.
*/
functor
export
   ControlFrame
import
   Tk
define
   /**
   % Gives a frame as well as an object. See the functor description above.
   */
   class ControlFrame from Tk.frame
      feat
	 simulator
	 nodelink
	 messagelog
	 playstopbutton
	 stepbutton
	 nodesbutton
	 logbutton
	 speedscale
	 clock
      attr
	 timeBetweenTicks
	 playing
	 nodesshowing
	 logshowing

      /**
      % Input arguments are the GraphicalSimulationEngine itself, a toplevel
      % for this frame, the two toplevels described above and some optional
      % selfdescribing init values.
      */
      meth init(GraphicalSimulationEngine
		Toplevel NodeLinkToplevel
		MessageLogToplevel
		play:Play<=false showNodeLink:InitShowNodes<=false
		showMessageLog:InitShowLog<=false speed:Speed<=100
		...)
	 Tk.frame, tkInit(parent:Toplevel)
	 self.simulator=GraphicalSimulationEngine

	 % set up control buttons 
	 self.playstopbutton={New Tk.button
			      tkInit(parent:self text:'Play'
				     action:self#play)}
	 self.stepbutton={New Tk.button
			    tkInit(parent:self text:'Step'
				   action: self#step)}
	 self.nodesbutton={New Tk.checkbutton
			  tkInit(parent:self text:"Show Node Layout"
				 action:self#ShowNodes)}
	 self.logbutton={New Tk.checkbutton
			 tkInit(parent:self text:"Show Message Log"
				action:self#ShowLog)}

	 % set up speedscale
	 self.speedscale={New Tk.scale
			  tkInit(parent:self orient:horizontal length:7#c
				 label: 'Delay between time ticks:'
				 'from':1000 to:0
				 action: self#setSpeed args:[int])}
	 {self setSpeed(Speed)}
	 
	 self.clock = {New Tk.label
		       tkInit(parent:self
			      text:'Time: '#{self.simulator getTime($)})}

	 {Tk.batch [grid(self.playstopbutton row:0 column:0 sticky:news)
		    grid(self.stepbutton row:0 column:1 sticky:nws)
		    grid(self.clock row:0 column:2 sticky:nes)
		    grid(self.speedscale row:1 column:0 columnspan:3
			 pady:8 sticky:ew)
		    grid(self.nodesbutton row:2 column:0 columnspan:2
			 sticky:nws)
		    grid(self.logbutton row:3 column:0 columnspan:2
			 sticky:nws)
		   ]}

	 % set up nodelink and message log to call hidenodes/hidelog when
	 % deleted
	 self.nodelink=NodeLinkToplevel
	 {Tk.send wm(protocol self.nodelink 'WM_DELETE_WINDOW'
		     {New Tk.action tkInit(parent:self.nodelink
					   action:self#hideNodes)})}
	 nodesshowing<-({Tk.returnAtom wm(state self.nodelink)}==normal)
	 % If the window is already visible, leave it otherwise check InitOpt
	 if @nodesshowing then
	    {self.nodesbutton tk(select)} 
	    {self.nodesbutton tkAction(action:self#HideNodes)}
	    {self.simulator nodeLinkActive(true)}
	 elseif InitShowNodes then
	    {self showNodes}
	 else
	    {self hideNodes}
	 end

	 self.messagelog=MessageLogToplevel
	 {Tk.send wm(protocol self.messagelog 'WM_DELETE_WINDOW'
		     {New Tk.action tkInit(parent:self.messagelog
					   action:self#hideLog)})}
	 logshowing<-({Tk.returnAtom wm(state self.messagelog)}==normal)
	 % If the window is already visible, leave it otherwise check InitOpt
	 if @logshowing then
	    {self.logbutton tk(select)}
	    {self.logbutton tkAction(action:self#HideLog)}
	    {self.simulator messageLogActive(true)}
	 elseif InitShowLog then
	    {self showLog}
	 else
	    {self hideLog}
	 end

         % Start a thread to handle play
	 if Play then {self play} end
	 thread {self PlayLoop} end
      end

      /**
      % Starts the simulation clock and makes steps happen automatically.
      */
      meth play
	 {self.stepbutton tk(configure state:disabled)}
	 {self.playstopbutton tk(configure text:'Stop')}
	 {self.playstopbutton tkAction(action:self#stop)}
	 @playing=true
      end

      meth PlayLoop
	 case @playing of true then % Will suspend if currently not playing
	    {self step}
	    {Delay @timeBetweenTicks}
	    {self PlayLoop}
	 [] done(W) then
	    W = unit
	 end
      end
      
      /**
      % Stops the simulation clock and automatic steps.
      */
      meth stop
	 playing<-_
	 {self.stepbutton tk(configure state:normal)}
	 {self.playstopbutton tk(configure text:'Play')}
	 {self.playstopbutton tkAction(action:self#play)}
      end

      /**
      % Do one step.
      % Should not be called when playing.
      */
      meth step
	 {self.simulator step}
	 {self.clock tk(configure text:'Time: '#{self.simulator getTime($)})}
      end

      /**
      % Sets the time to be waited between every discrete time event.
      */
      meth setSpeed(Speed)
	 {self.speedscale tk(set Speed)}
	 timeBetweenTicks<-Speed
      end

      /**
      % Displays the NodeLinkToplevel if not allready visible.
      */
      meth showNodes
	 if {Not nodesshowing<-true} then
	    {self.nodesbutton tk(invoke)}
	 end
      end
      
      meth ShowNodes
	 nodesshowing<-true
	 {self.nodesbutton tkAction(action:self#HideNodes)}
	 {self.simulator nodeLinkActive(true)}
	 {Tk.send wm(deiconify self.nodelink)}
      end

      /**
      % Hides the NodeLinkToplevel if visible.
      */
      meth hideNodes
	 if nodesshowing<-false then
	    {self.nodesbutton tk(invoke)}
	 end
      end
      
      meth HideNodes
	 nodesshowing<-false
	 {self.nodesbutton tkAction(action:self#ShowNodes)}
	 {self.simulator nodeLinkActive(false)}
	 {Tk.send wm(withdraw self.nodelink)}
      end

      /**
      % Displays the MessageLogToplevel if not allready visible.
      */
      meth showLog
	 if {Not logshowing<-true} then
	    {self.logbutton tk(invoke)}
	 end
      end
      
      meth ShowLog
	 logshowing<-true
	 {self.logbutton tkAction(action:self#HideLog)}
	 {self.simulator messageLogActive(true)}
	 {Tk.send wm(deiconify self.messagelog)}
      end

      /**
      % Hides the MessageLogToplevel if visible.
      */
      meth hideLog
	 if logshowing<-false then
	    {self.logbutton tk(invoke)}
	 end
      end
      
      meth HideLog
	 logshowing<-false
	 {self.logbutton tkAction(action:self#ShowLog)}
	 {self.simulator messageLogActive(false)}
	 {Tk.send wm(withdraw self.messagelog)}
      end

      /**
      % Makes sure the playthread stops and then closes the two toplevels.
      % Will not work if a step has suspended.
      */
      meth tkClose
	 if {IsDet @playing} then W in
	    playing <- done(W)
	    {Wait W}
	 end
	 {self.nodelink tkClose}
	 {self.messagelog tkClose}
      end
   end
end

	 
