/**
The Simulation engine is a simple engine that maintains nodes and links
its own time. (There is no graphics involved.)
Time progresses when the engine gets a step call.
Nodes are created with a listener. On every step the listener will be
invoked with a message about the current time and a list of any incomming
messages or notifications of lost links.
Links are created by specifying a link id, a from and a to node plus
specific link behaviour. To use the link send is called with the linkid and
a message of any format. Specific link behaviour is how many slots the link
should have and how messages should be shifted in these slots on each step.
Default behaviour is FIFO moving messages one slot forward on each step.

On each step the folloing happens in order:
1. The time is progressed
2. All links are traversed and shifted
3. All nodes are traversed and their listeners are informed of
   the step and any messages.
*/
functor
export
   SimulationEngine
import
   Utils(takeDet:TakeDet assert:Assert
	 mapMethod:MapMethod allMeth:AllMeth) at '../Utils.ozf'
define
   class SimulationEngine
      feat
	 Nodes
	 Links
      attr
	 time
      meth init
	 self.Nodes={NewDictionary}
	 self.Links={NewDictionary}
	 time<-0
      end

      /**
      % progresses time one step
      */
      meth step O N in
	 O=time<-N
	 N=O+1
	 {AllMeth {Dictionary.items self.Links} step}
	 {AllMeth {Dictionary.items self.Nodes} step(N)}
      end

      /**
      % creates a node with the specified id and
      % listener. On each step the listener will be
      % invoked as {Listener step(Time Msgs)}. Where
      % Time is an integer and Msgs is a list of
      % message#FromNodeId or notifications of lost
      % links.
      % The message format is whatever has been used
      % by sender and shifter. A notification of a
      % lost link is connectionLost#FromNodeId.
      */
      meth createNode(NodeId Listener)
	 if {Dictionary.member self.Nodes NodeId} then
	    raise nodeAlreadyExists(NodeId) end
	 else
	    self.Nodes.NodeId:={New Node init(NodeId Listener)}
	 end
      end

      /**
      % removes the node and no further messages will
      % be passed to Listener.
      */
      meth removeNode(NodeId)
	 RmNode=self.Nodes.NodeId
      in
	 if {Not {Dictionary.member self.Nodes NodeId}} then
	    raise nodeNotFound(NodeId) end
	 end
	 {Dictionary.remove self.Nodes NodeId}
	 {RmNode crash}
      end

      /**
      % creates a link with LinkId between the two
      % specified nodes that must exist. A number
      % of slots and a Shifter (possibly default)
      % must be specified.
      % If a Shifter is specified it will be invoked as
      {Shifter GetInMsg SlotArray NofSlots ?OutMsg}
      % on each step. GetInMsg is a function that on
      % invocation returns the next message about to
      % enter the link or nil if none present.
      % SlotArray is an array with one entry for each
      % slot in the link. An entry may be a message
      % ore nil. NofSlots is the length of SlotArray
      % provided for convenience. Shifter should
      % return any message that leaves the link and
      % should arrive to the node in OutMsg.
      % Shifter may not alter any messages, only
      % move or possibly loose them.
      */
      meth createLink(LinkId FromNodeId ToNodeId NofSlots Shifter)
	 FromNode=self.Nodes.FromNodeId
	 ToNode=self.Nodes.ToNodeId
      in
	 if {Dictionary.member self.Links LinkId} then
	    raise linkAlreadyExists(LinkId) end
	 end
	 self.Links.LinkId:={New Link init(LinkId FromNode ToNode
					   NofSlots Shifter)}
      end

      /**
      % removes a link in one of two ways specified
      % by How. How may be crash or flush. crash
      % means remove immediately and loose any
      % messages. flush means close link now, but
      % allow any messages present int the link to
      % arrive to the destination before removing it.
      % A unary procedure DoneAlert will be invoked
      % when the link is removed (immediatly for crash).
      */
      meth removeLink(L How DoneAlert)
	 proc{Done} {Dictionary.remove self.Links L} {DoneAlert} end
      in
	 if {Not {Dictionary.member self.Links L}} then
	    raise linkNotFound(L) end
	 end
	 {self.Links.L remove(How Done)}
      end

      /**
      % sends Msg on link with id L.
      */
      meth send(L Msg)
	 if {Not {Dictionary.member self.Links L}} then
	    raise linkNotFound(L) end
	 end
	 {self.Links.L send(Msg)}
      end
      
      /**
      % returns the current state of all links as
      % a list of state(id:LinkId
      %                 messages:CopyofMessageArray
      %                 flushing:IsFlushing)
      % where IsFlushing is true after removeLink has
      % been called with flush and before DoneAlert
      % has been called (see removeLink).
      */
      meth getLinkStates($)
	 {MapMethod {Dictionary.items self.Links} getLinkState}
      end

      /**
      % returns the current time.
      */
      meth getTime($)
	 @time
      end
   end

   % Nodes are the base for an executing "machine"
   % This class will notify the owner of steps with the current time
   % and any incomming messages
   class Node
      feat
	 id
	 listener
      attr
	 instream
	 instreamend
	 crashed
	 
      meth init(Id Listener)
	 self.id=Id
	 self.listener=Listener
	 instream<-_
	 instreamend<-@instream
	 crashed<-false
      end

      meth step(Time)
	 {Assert {Not @crashed}#"Crashed node got step"}
	 Msgs={self GetReceived($)}
      in
	 {self.listener step(Time Msgs)}
      end
      
      % Extract the determined part of @instream and remember
      % where to start next time
      meth GetReceived(?Msgs)
	 N O=instream<-N
      in
	 {TakeDet O N Msgs}
      end

      % Interface method for Links to tell about received messages
      meth receive(Msg FromId)
	 if {Not @crashed} then
	    NEnd OEnd=instreamend<-NEnd
	 in
	    OEnd=(Msg#FromId)|NEnd
	 end
      end

      meth crash
	 crashed<-true
      end
   end

   % Links are the base for communication
   % They are created by specifying 2 node ids to the engine
   class Link
      feat
	 id
	 nofslots
	 slots
	 fromnode
	 tonode
      attr
	 fromstream
	 fromstreamend
	 flushing
	 shifter
	 
      meth init(Id FromNode ToNode NofSlots InShifter)
	 self.id=Id
	 self.nofslots=NofSlots
	 self.slots={NewArray 1 NofSlots nil}
	 self.fromnode=FromNode
	 self.tonode=ToNode
	 fromstreamend<-@fromstream
	 flushing<-false#proc{$} skip end
	 case InShifter of default then
	    shifter<-Shifter
	 else
	    {Assert {IsProcedure InShifter}#"Expected procedure for Shifter"}
	    shifter<-InShifter
	 end
      end

      meth send(Msg)
	 {Assert {Not @flushing.1}#'Message sent on link scheduled for removal'}
	 NEnd OEnd=fromstreamend<-NEnd
      in
	 OEnd=Msg|NEnd
      end
      
      meth step
	 fun{GetInMsg}
	    if {IsDet @fromstream} then
	       N O=fromstream<-N
	    in
	       N=O.2
	       O.1
	    else
	       nil
	    end
	 end
	 OutMsg
      in
	 % Let shifter shift
	 {@shifter GetInMsg self.slots self.nofslots ?OutMsg}

	 % Take care of any OutMsg
	 if {IsDet OutMsg} andthen OutMsg \= nil then
	    {self.tonode receive(OutMsg self.fromnode.id)}
	 end

	 % If we are flushing and now empty, "loose" the connection
	 if @flushing.1 andthen {Not {IsDet @fromstream}} then Found in
	    for N in 1..self.nofslots do
	       if self.slots.N \= nil then Found=true end
	    end
	    if {Not {IsDet Found}} then
	       {self.tonode receive(connectionLost self.fromnode.id)}
	       {@flushing.2} % Report done
	    end
	 end
      end
      
      meth remove(How Done)
	 case How of crash then
	    {self.tonode receive(connectionLost self.fromnode.id)}
	    {Done}
	 [] flush then
	    flushing<-true#Done
	 end
      end

      meth getLinkState($)
	 state(id:self.id
	       messages:{Array.clone self.slots} flushing:@flushing.1)
      end
   end

   % A shifter is responsible for moving the messages in the SlotArray.
   % If a (to the link) incoming message is needed it can be retrieved with
   % the provideed GetInMsg procedure. If a message is considerred to have
   % arrived to the destination it shall be put in OutMsg. Undetermined OutMsg
   % or nil will be interpreted as no message arrived.
   %
   % This default shifter moves every message one step forward.
   proc{Shifter GetInMsg SlotArray NofSlots ?OutMsg}
      % Last msg arrives
      OutMsg = SlotArray.NofSlots

      % Messages inside SlotArray step one step forward
      {For NofSlots 2 ~1
       proc{$ N}
	  SlotArray.N:=SlotArray.(N-1)
       end}

      % One incomming message is moved into SlotArray
      SlotArray.1:={GetInMsg}
   end
end






